% This is the main run file for estimating the VECM-SMSS model with a full
% covariance for the state transitions.  Various settings at the beginning
% of this script control its execution.  Note that there are more settings
% (and possibilities) here than what is actually used in the paper.  Some
% of these, however, are experimental and may result in errors.
%
% NOTICE:
% -------
% Permission is hereby granted to use this code freely for academic
% purposes only, provided that the paper is duly cited as:
%
% Eisenstat, E., Chan, J. C. C. and R. W. Strachan (2015). "Stochastic Model
% Specification Search for Time-Varying Parameter VARs," Econometric
% Reviews, forthcoming.
%
% Any other use of this code, particularly for comercial purposes, is
% strictly prohibited without the prior written consent of the authors.
% Furthermore, this code comes without technical support of any kind.  It
% is expected to reproduce the results reported in the paper.  However,
% beware that the notation may not the match text.  Likewise, it is not
% gauranteed to work under different hardware and / or software settings,
% other than those used for its original design and implementation.
%
% Under no circumstances will the authors be held responsible for any use
% (or misuse) of this code in any way.  If you do not agree to the above,
% you do not have permission to use this code.

clear;
scale_data = 1; % standardize all series to have std. dev. of 1
center_data = 1; % standardize all series to have mean 0

% algorithm-specific controls
rand_mu = 0; % sample mu randomly (experimental)
agjoint = 0; % sample alpha, gamma jointly (experimental)
lasso_alpha = 0; % use a lasso prior on the variance of alpha
lasso_Phi = 1; % use a lasso prior on the state covariances
do_expansion = 0; % use parameter expansion on om, Phi, gamma
do_ishift = [1 1]; % do a distn-invariant translation / scaling of draws

nsims = 25000; % desired number of MCMC simulations
burnin = .1 * nsims; % burn-in simulations to discard
progstep = 100; % frequency of progress updates
                % (i.e. every "progstep" simulations)

% set this to save every "simstep"^th draw;
% useful for running long chains on
% windows machines with limited memory
simstep = 5;
svsims = floor( nsims / simstep );

p = 3; % number of AR lags

% process data
load BP2002data_v19.csv;
data = BP2002data_v19;
Ydata = data( :, 1:3 );

[T n] = size( Ydata );
dscale = 1 + scale_data * ( std( Ydata, 1 ) - 1 );
dcenter = center_data * mean( Ydata, 1 );
Ydata = ( Ydata  - repmat( dcenter, T, 1 ) ) ./ repmat( dscale, T, 1 );

Y0 = Ydata( 1:p, : )';  % store the first p obs as init cond
shortY = Ydata( ( p + 1 ):end, : )';

T = T - p;
y = shortY( : );

X0 = zeros( n * p, T );
for j = 1:p
    X0( ( j - 1 ) * n + 1:j * n, : ) = [Y0( :, p - j + 1:end ) shortY( :, 1:T - j )];
end

%%% co-integration terms (tax - spend, spend - gdp) %%%
surp = data( ( p + 1 ):end, end - 1:end );
dscale_surp = 1 + 0*scale_data * ( std( surp, 1 ) - 1 ); % experimental
dcenter_surp = 0*center_data * mean( surp, 1 ); % experimental
surp = ( surp - repmat( dcenter_surp, T, 1 ) ) ./ repmat( dscale_surp, T, 1 );
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

Xdata = [ones( 1, T ); data( ( p + 1 ):end, 4:( end - 2 ) )'; surp'];
X0 = [Xdata; X0];

% define useful constants
k = n * p + size( Xdata, 1 );
nk = 0.5 * n * ( n - 1 ) + n * k; % size of parameters vector

% construct the big X
X = X0;
cidx = zeros( k, 1 );
for i = 1:n - 1
    X = [X; -shortY( 1:i, : ); X0];
    cidx = [cidx; ones( i, 1 ); zeros( k, 1 )];
end

% indices to recover covariance coefficients
cidx = logical( cidx );
bidx = logical( 1 - cidx );

% construct prototype F, G, Z, W
z = cell( n, 1 );
for i = 1:n
    z{i} = ones( k + i - 1, 1 );
end
Z = sparse( repmat( blkdiag( z{:} ), 1, T ) );
W = kron( speye( T ), blkdiag( z{:} )' ); Widx = logical( W );
G = Z; Gidx = logical( G ); % note: this is actually G' in the paper
Z( Z ~= 0 ) = X( : ); Z = Z';

F = [];
for j = 1:nk - 1
    F = blkdiag( F, ones( j, 1 ) );
end
F = repmat( [sparse( size( F, 1 ), 1 ) F], 1, T ); % this will hold the
                                                   % gammas for sampling Phi
% some useful indices
Fidx = logical( F );
uidx = triu( ones( nk, nk ) ); uidx( :, end ) = 0;
biguidx = logical( repmat( uidx, 1, T ) );
Phi_idx = logical( triu( ones( nk, nk ) ) - eye( nk ) );

% priors
% for the IW prior set (used here):
s0h = 14; S0h = 0.01 ^ 2 * ( s0h - n - 1 ) * eye( n );

% for the IG priors, the following corresponds to the marginal prior on
% Sig_h(i,i) under the inverse wishart (used in the diagonal transition cov
% version):
s0hIG = ( s0h - n + 1 ) / 2; S0hIG = diag( S0h ) / 2;

% the rest are standard
a0 = zeros( nk, 1 ); A0 = eye( nk );
phi0 = zeros( nk * ( nk - 1 ) / 2, 1 ); Phi0 = eye( length( phi0 ) ) * 10;
Hsmall = spdiags( -ones( T, 1 ), - 1, speye( T ) );
H = spdiags( -ones( T * nk, 1 ), -nk, speye( T * nk ) );
HH = H' * H; K = blkdiag( A0, HH );
G0 = ones( nk, 1 );
mu0 = zeros( nk, 1 ); M0 = ones( nk, 1 );
lam20( 1 ) = 0.1; lam20( 2 ) = 0.1; % the same for all lasso's

% starting values
[~, Sig] = ldl( cov( Ydata ) );
h = repmat( log( diag( Sig ) ), T, 1 );
Sigh = S0h / ( s0h + n + 1 ); % start with the prior mode
bigSig = spdiags( exp( -h ), 0, speye( n * T ) );
tau2 = exprnd( 1, nk, 1 );
mu = zeros( nk, 1 );
om_st = sqrt( tau2 ) .* randn( nk, 1 );
om = om_st .* ( om_st > 0 );
alpha = randn( nk, 1 );
gamma = randn( nk, T );
Phi = speye( nk );
scl2_1 = ones( nk, 1 );
scl2_2 = 1;
loc = zeros( nk, 1 );

% allocate space for draws
s_alpha = zeros( nk, svsims );
s_gamma = zeros( nk, T, svsims );
s_beta = zeros( nk, T, svsims );
s_Phi = zeros( nk, nk, svsims );
s_Sig = zeros( n, T, svsims );
s_Sigh = zeros( n, n, svsims );
s_lam2 = zeros( 1, svsims );
s_tau2 = zeros( nk, svsims );
s_mu = zeros( nk, svsims );
s_om_st = zeros( nk, svsims );
s_om = zeros( nk, svsims );
s_adj = zeros( nk, 3, svsims );

% final initializations
disp( ['Sampling ' num2str( burnin + nsims ) ' draws from the posterior...'] );

% the following is needed ONLY on Windows
% to display progress properly; on Linux
% it has no effect (but should be left defined)
txtlen = 0;

% initialize the random number generator
rand( 'state', sum( 100 * clock ) );
randn( 'state', sum( 200 * clock ) );

tic; % start the clock
for isim = 1:( burnin + nsims )

    W( Widx ) = vec( X .* repmat( om, 1, T ) );
    WPhi = W * kron( speye( T ), Phi' );
    
    if agjoint
        % sample alpha, gamma jointly
        ZWSig = [Z WPhi]' * bigSig;
        Delta_hat = K + ZWSig * [Z WPhi];
    
        % chol() is way to slow here...
        % this implementation is identical to
        % sampling alpha followed by gamma | alpha
        C = mydecomp( Delta_hat, nk ); % C' * C = Delta_hat
        delta_hat = C \ ( C' \ ( K * [a0; zeros( T * nk, 1 )] + ZWSig * y ) );
        delta = delta_hat + C \ randn( ( T + 1 ) * nk, 1 );

        alpha = delta( 1:nk );
        gamma = reshape( delta( ( nk + 1:end ) ), nk, T );
    else
        % sample alpha | gamma
        A_hat = A0 + Z' * bigSig * Z;
        a_hat = A_hat \ ( A0 * a0 + Z' * bigSig * ( y - WPhi * gamma( : ) ) );
        alpha = a_hat + chol( A_hat ) \ randn( nk, 1 );
    
        % sample gamma | alpha
        Gam_hat = HH + WPhi' * bigSig * WPhi;
        gam_hat = reshape( Gam_hat \ ( WPhi' * bigSig * ( y - Z * alpha ) ), nk, T );
        gamma = gam_hat + reshape( chol( Gam_hat ) \ randn( T * nk, 1 ), nk, T );
    end
    
    if lasso_alpha
    % do lasso on A0
        lam2_a = gamrnd( lam20( 1 ) + nk, 1 / ( lam20( 2 ) + sum( 1 ./ diag( A0 ) ) / 2 ) );
        A0 = diag( igrand( sqrt( lam2_a ) ./ abs( alpha - a0 ), lam2_a ) );
    end

    % sample Sig_t
    err = y - [Z WPhi] * [alpha; gamma( : )];
    h = MVSVRW( log( err .^ 2 + 0.001 ), h, inv( Sigh ), speye( n ) );
    bigSig = spdiags( exp( -h ), 0, speye( n * T ) );
    
    shorth = reshape( h, n, T );
    errh = shorth( :, 2:end ) - shorth( :, 1:end - 1 );
    sseh = errh * errh';
    Sigh = iwishrnd( S0h + sseh, s0h + T - 1 ); % correlated volatilities
    
    % sample om, om_st, tau2, lam2
    G( Gidx ) = vec( X .* ( Phi' * gamma ) );
    [om, om_st, tau2, lam2] = get_om_tau2_lam2( lam20, y - Z * alpha, G', bigSig, om, tau2, mu );
  
    if rand_mu
        % sample mu
        M_hat = 1 ./ ( M0 + 1 ./ tau2 );
        mu_hat = M_hat .* ( M0 .* mu0 + om_st ./ tau2 );
        mu = mu_hat + sqrt( M_hat ) .* randn( nk, 1 );
        lam2_mu = gamrnd( lam20( 1 ) + nk, 1 / ( lam20( 2 ) + sum( 1 ./ M0 ) / 2 ) );
        M0 = igrand( sqrt( lam2_mu ) ./ abs( mu - mu0 ), lam2_mu );
    end
    
    % sample Phi
    % the conditional model is: y = Z*alpha + W*gamma + W*F'*phi + e
    bigGam = reshape( repmat( gamma, nk, 1 ), nk, nk * T );
    F( Fidx ) = bigGam( biguidx );
    err = y - [Z W] * [alpha; gamma( : )];
    WF = W * F';
    
    Phi_hat = Phi0 + WF' * bigSig * WF;
    phi_hat = Phi_hat \ ( Phi0 * phi0 + WF' * bigSig * err );
    Phi( Phi_idx ) = phi_hat + chol( Phi_hat ) \ randn( length( phi_hat ), 1 );

    if lasso_Phi
    % do lasso on Phi0
        lam2_f = gamrnd( lam20( 1 ) + length( phi0 ), 1 / ( lam20( 2 ) + sum( 1 ./ diag( Phi0 ) ) / 2 ) );
        Phi0 = diag( igrand( sqrt( lam2_f ) ./ abs( Phi( Phi_idx ) - phi0 ), lam2_f ) );
    end
    
    if do_expansion
        % sample the expansion parameters
        sseg = sum( ( Hsmall * gamma' ) .^ 2, 1 )';
        G0 = gamrnd( ( 1 + T ) / 2, 2 ./ ( n * T + sseg ) );
        
        longG0 = repmat( sqrt( G0 ), T, 1 );
        HG0 = spdiags( [-longG0 longG0], [-nk 0], T * nk, T * nk );
        HH = HG0' * HG0;
        
        if agjoint
            K = blkdiag( A0, HH );
        end
    end
    
    if do_ishift( 1 )
        % apply one distn-invariant scale transformation
        % to lam2 & tau2 (e.g. Liu and Sabatti, 2000)
        scl2_2 = gamrnd( lam20( 1 ) + nk / 2, 1 / ( lam20( 2 ) * lam2 + sum( ( om_st - mu ) .^ 2 ./ tau2 ) / 2 ) );
        tau2 = tau2 / scl2_2;
        
        if rand_mu
            % one more time for lam2_mu, M0
            scl2_2m = gamrnd( lam20( 1 ) + nk / 2, 1 / ( lam20( 2 ) * lam2_mu + sum( ( mu - mu0 ) .^ 2 .* M0 ) / 2 ) );
            M0 = M0 * scl2_2m;
        end
        
        if lasso_alpha
        % one more for lam2_a, A0
            scl2_2a = gamrnd( lam20( 1 ) + nk / 2, 1 / ( lam20( 2 ) * lam2_a + sum( ( alpha - a0 ) .^ 2 .* diag( A0 ) ) / 2 ) );
            A0 = A0 * scl2_2a;
        end
        
        if lasso_Phi
        % one more for lam2_f, Phi0
            scl2_2f = gamrnd( lam20( 1 ) + length( phi0 ) / 2, 1 / ( lam20( 2 ) * lam2_f + sum( ( Phi( Phi_idx ) - phi0 ) .^ 2 .* diag( Phi0 ) ) / 2 ) );
            Phi0 = Phi0 * scl2_2f;
        end
    end
    
    if do_ishift( 2 )
        % apply j distn-invariant location transformations
        % to alpha and gamma (e.g. Liu and Sabatti, 2000)
        Om = diag( om ) * Phi';
        Loc_hat = speye( nk ) + Om' * A0 * Om;
        loc_hat = Loc_hat \ ( gamma( :, 1 ) - Om' * A0 * ( alpha - a0 ) );
        loc = loc_hat + chol( Loc_hat ) \ randn( nk, 1 );
        alpha = alpha + om .* loc;
        gamma = gamma - repmat( loc, 1, T );
    end

    % save draws
    if isim > burnin && mod( isim - burnin, simstep ) == 0
        isave = ( isim - burnin ) / simstep;
        s_alpha( :, isave ) = alpha;
        s_gamma( :, :, isave ) = diag( sqrt( G0 ) ) * gamma;
        s_beta( :, :, isave ) = repmat( alpha, 1, T ) + diag( om ) * Phi' * gamma;
        s_Phi( :, :, isave ) = diag( sqrt( G0 ) ) * Phi' * diag( sqrt(  1./ G0 ) );
        s_Sig( :, :, isave ) = exp( shorth ); %Sig;
        s_Sigh( :, :, isave ) = Sigh;
        s_lam2( :, isave ) = lam2;
        s_tau2( :, isave ) = tau2;
        s_mu( :, isave ) = mu;
        s_om_st( :, isave ) = om_st .* sqrt(  1./ G0 );
        s_om( :, isave ) = om .* sqrt(  1./ G0 );
        s_adj( :, :, isave ) = [sqrt( scl2_1 ) [sqrt( scl2_2 ); zeros( nk - 1, 1 )] loc];
    end

    % show progress
    txtlen = showprog( isim, burnin + nsims, progstep, txtlen );
end
fprintf( '\nSampling completed after %5.3f minutes.\n', toc / 60 );

figure;
mbeta0 = mean( s_beta, 3 )'; mbeta = zeros( size( mbeta0 ) );
mbeta( :, 1:nnz( cidx ) ) = mbeta0( :, cidx );
mbeta( :, nnz( cidx ) + 1:end ) = mbeta0( :, bidx );

%%% ANALYSIS
% some diagnostics
% do some MCMC mixing analysis and plot some beta's
ef_om_st = ineff_factor( s_om_st );
ef_beta = ineff_factor( reshape( s_beta, nk * T, svsims ) );
boxplot( [ef_om_st; ef_beta], [repmat( { 'omega_star' }, nk, 1 ); repmat( { 'beta' }, nk * T, 1 )] );
figure, plot( max( mbeta, [], 1 ) - min( mbeta, [], 1 ), ':*' );

% plot the time ivariant / constant param probabilities
prtiv0 = sum( s_om == 0, 2 )' / svsims; prtiv = zeros( size( prtiv0 ) );
prtiv( 1:nnz( cidx ) ) = prtiv0( cidx );
prtiv( nnz( cidx ) + 1:end ) = prtiv0( bidx );
figure, plot( 1:size( s_om, 1 ), prtiv, ':*' )

% impulse responses
scale_adj = dscale' * ( 1 ./ dscale );
per = 1959.5:.25:2011.75;
t_start = find( per == 2000 );
s = 20;
[ab2, c, structvars, IR, err] = ir_BP2002_vecm_sv( s_beta, s_Sig, cidx, t_start, s, 2.08 * dscale( 3 ) / dscale( 1 ), p, diag( 1 ./ dscale_surp ) * [1 -1 0; 0 1 -1] * diag( dscale ) );
nerr = nnz( err );
bands = floor( [.16 .5 .84] * ( svsims - nerr ) );
IRsort = sort( IR( :, :, :, ~err ), 4 );
make_shaded_IR_plots;